from typing import Dict, List, Union, Tuple
import json
import os
from uuid import uuid4
from datetime import date
from .lectric_types import HashAlgo
from requests import Response
from functools import wraps

from .models.input_data import InputData
from .models.index_in_spec import IndexInSpec
from .models.collection import Collection
from .models.collection_in_spec import CollectionInSpec
from .models.query_spec import QuerySpec
from .models.vector_query_spec import VectorQuerySpec
from .models.query_response import QueryResponse
from .models.collection_metadata import CollectionMetadata
from .models.exact_collection_entry import ExactCollectionEntry
from .models.schema_info_exact_schema_info_get_response_schema_info_exact_schema_info_get import SchemaInfoExactSchemaInfoGetResponseSchemaInfoExactSchemaInfoGet

from .helpers import *
from .client import AuthenticatedClient
from .types import File as fastAPIFile

# defaults
from .api.default import  root_get, root_auth_auth_get, list_connections_connections_get

# collections
from .api.collection import (
    create_collection_collection_create_post,
    exists_collection_collection_exists_name_get,
    is_empty_collection_collection_empty_name_get,
    get_collection_collection_get_name_get,
    list_collections_collection_list_get,
    get_indexes_collection_indexes_name_get,
    drop_collection_collection_delete,
    sizeof_collection_collection_size_name_get,
    delete_entities_collection_entities_delete,
    sample_collection_sample_get,
    get_url_collection_get_url_get
)

# ingest
from .api.ingest import ingest_ingest_post
from .api.ingest import ingest_w_file_ingest_file_post

# index
from .api.index import create_index_create_post, delete_index_collection_name_field_name_delete

# query
from .api.query import query_query_fields_post, query_query_vectors_post

# exact
from .api.exact import (
    create_collection_exact_create_collection_post,
    put_entry_by_uri_exact_hash_collection_name_uri_put,
    put_entry_by_file_exact_hash_collection_name_file_put,
    lookup_by_uri_exact_lookup_hash_uri_collection_name_get,
    lookup_by_file_exact_lookup_file_collection_name_post,
    list_collections_exact_list_get,
    get_available_algos_exact_algos_get,
    is_algo_available_exact_available_algo_get,
    info_exact_info_get,
    has_revlookup_exact_revlookup_get,
    revlookup_list_exact_revlookup_list_get,
    has_dupecounter_exact_dupecounter_get,
    dupes_list_exact_counter_dupes_get,
    raw_query_exact_experimental_raw_query_post,
    ping_exact_ping_post,
    schema_info_exact_schema_info_get,
    hard_delete_collection_exact_collection_hard_delete_delete,
    hard_delete_entries_by_file_exact_entries_by_file_hard_delete_delete,
    hard_delete_entries_exact_entries_hard_delete_delete,
    soft_delete_collection_exact_collection_soft_delete_delete,
    soft_delete_entries_by_file_exact_entries_by_file_soft_delete_delete,
    soft_delete_entries_exact_entries_soft_delete_delete,
    select_entries_exact_select_post,
    get_collection_metadata_exact_collection_metadata_get,
    restore_entries_exact_entries_restore_patch,
    restore_collection_exact_collection_restore_patch,
    trashcan_entries_exact_trashcan_entries_get,
    get_exact_collection_entries_exact_collection_entries_get
)

from .models.body_put_entry_by_file_exact_hash_collection_name_file_put import BodyPutEntryByFileExactHashCollectionNameFilePut
from .models.body_lookup_by_file_exact_lookup_file_collection_name_post import BodyLookupByFileExactLookupFileCollectionNamePost
from .models.body_ingest_w_file_ingest_file_post import BodyIngestWFileIngestFilePost
from .models.body_hard_delete_entries_by_file_exact_entries_by_file_hard_delete_delete import BodyHardDeleteEntriesByFileExactEntriesByFileHardDeleteDelete
from .models.body_hard_delete_entries_exact_entries_hard_delete_delete import BodyHardDeleteEntriesExactEntriesHardDeleteDelete
from .models.body_soft_delete_entries_by_file_exact_entries_by_file_soft_delete_delete import BodySoftDeleteEntriesByFileExactEntriesByFileSoftDeleteDelete
from .models.body_soft_delete_entries_exact_entries_soft_delete_delete import BodySoftDeleteEntriesExactEntriesSoftDeleteDelete

def propagate_exception(func):
    @wraps(func)
    def propagator(*args, **kwargs):

        response = func(*args, **kwargs)

        if response:
            if response.status_code == HTTPStatus.OK:
                return response.parsed
            else:
                raise RuntimeError(f"HTTP Code: {response.status_code}, {response}")
        else:
            raise RuntimeError("No response")

    return propagator


class LectricClient:

    def __init__(self, api_url: str, api_key: str=None,
                    cookies: Dict[str, str]={}, timeout:int=60000) -> None:
        """The Lectric client class

        Args:
            api_url (str): The base url for the api service
            api_key (str, optional): The API key for the service. Defaults to None.
            cookies (Dict[str, str], optional): Optional cookies to be maintained. Defaults to {}.
            timeout (int, optional): Request timeout. Defaults to 60000.
        """
        self.api_url = api_url

        self.client = AuthenticatedClient(base_url=self.api_url,
                                    token=api_key, cookies=cookies,
                                    headers={"X-API-KEY": api_key} if api_key else {},
                                    timeout=timeout)

        response_dict = json.loads(self.verify_connection(api_key is not None).content)

        self.api_build_id = response_dict.get("lectric-build-id")
        self.backend_db = response_dict.get("lectric-vdb-backend")

    def verify_connection(self, is_authenticated: bool=False) -> Response:
        """Test the connection to the Lectric server

        Returns:
            bool|RuntimeError: Either True for success or failure
        """

        if is_authenticated:
            response = root_auth_auth_get.sync_detailed(client=self.client)
        else:
            response = root_get.sync_detailed(client=self.client)

        if response.status_code != HTTPStatus.OK.value:
            raise RuntimeError(f"HTTP Code: {response.status_code}, {response}")

        return response

    @propagate_exception
    def list_connections(self) -> List[str]:
        """List the connections created within the vector db

        Returns:
            List[str]: A list of the connection names
        """
        return list_connections_connections_get.sync_detailed(client=self.client)


    ########################## collections ###############################

    @propagate_exception
    def create_collection(self, coll_in_spec: CollectionInSpec,
                                    with_reverse_lookup: bool=False,
                                    with_dupe_counter: bool=False,
                                    hash_algo: HashAlgo =HashAlgo.md5
                                    ) -> Collection:
        """Create an exact OR vector collection. If `coll_in_spec.approx == True` then the collection created will be a vector/approximate collection; if not it will be an exact one

        Args:
            coll_in_spec (CollectionInSpec): The collection specification to be created
            with_reverse_lookup (bool): Only relevant to exact collections. Creates a sidecar collection to do reverse lookups. Defaults to False to False.
            with_dupe_counter (bool): Only relevant to exact collections. Adds a counting attribute to each member of the collection for tracking and counting.
            hash_algo (HashAlgo): The hash algorithm to use for exact collections. Default is HashAlgo.md5

        Returns:
            Collection: An object representing the created collection
        """
        if coll_in_spec.approx:
            return create_collection_collection_create_post.sync_detailed(client=self.client, body=coll_in_spec)
        return create_collection_exact_create_collection_post.sync_detailed(client=self.client,
                                                                                body=coll_in_spec,
                                                                                with_reverse_lookup=with_reverse_lookup,
                                                                                with_dupe_counter=with_dupe_counter,
                                                                                hash_algo=hash_algo.name)


    @propagate_exception
    def collection_exists(self, name: str, exact: bool=False) -> bool:
        """Check if a collection exists

        Args:
            name (str): The name of the collection
            exact (bool): Looking for an exact collection?

        Returns:
            bool: True if the collection exist, else False
        """
        return exists_collection_collection_exists_name_get.sync_detailed(client=self.client, name=name, exact=exact)


    @propagate_exception
    def is_collection_empty(self, name: str) -> bool:
        """Check if the collection has data or is empty

        Args:
            name (str): The name of the collection

        Returns:
            bool: True if the collection is empty, else False
        """
        return is_empty_collection_collection_empty_name_get.sync_detailed(client=self.client, name=name)


    @propagate_exception
    def get_collection(self, name: str) -> Collection:
        """Get a collection by name. Each collection has a unique, user defined name

        Args:
            name (str): The name of the collection.

        Returns:
            Collection: A representation of the collection
        """
        return get_collection_collection_get_name_get.sync_detailed(client=self.client, name=name)


    @propagate_exception
    def get_indexes(self, collection_name: str) -> List[str]:
        """Get a list of the field names for which indexes are defined

        Args:
            collection_name (str): The name of the collection for which indexes are queried

        Returns:
            List[str]: A list of the field names for which indexes are defined
        """
        return get_indexes_collection_indexes_name_get.sync_detailed(client=self.client, name=collection_name)


    @propagate_exception
    def list_collections(self, exact_only: bool=False) -> List[str]:
        """List the names of the collections

        Returns:
            List[str]: The collection names
        """
        if exact_only:
            res = list_collections_exact_list_get.sync_detailed(client=self.client)
            return res
        return list_collections_collection_list_get.sync_detailed(client=self.client)


    @propagate_exception
    def hard_drop_collection(self, collection_name: str, exact: bool=False) -> None:
        """Drop (delete) a collection by name

        Args:
            name (str): The name of the collection to be dropped
            exact (bool): The collection is exact

        Returns:
            _type_: None
        """
        if not exact:
            return drop_collection_collection_delete.sync_detailed(client=self.client, name=collection_name)
        else:
            return hard_delete_collection_exact_collection_hard_delete_delete.sync_detailed(client=self.client, collection_name=collection_name)

    @propagate_exception
    def soft_drop_collection(self, collection_name: str, exact: bool=False, deleter: str=None) -> None:
        """Drop (delete) a collection by name

        Args:
            name (str): The name of the collection to be dropped
            exact (bool): The collection is exact

        Returns:
            _type_: None
        """
        if not exact:
            raise NotImplementedError("Soft-delete entries has not been implemented for approx yet")
        else:
            return soft_delete_collection_exact_collection_soft_delete_delete.sync_detailed(client=self.client, collection_name=collection_name, deleter=deleter)

    @propagate_exception
    def soft_delete_entities(self, collection_name: str,
                ids: Union[List[str], List[int]]=None,
                urls: List[str]=None,
                file: Union[BinaryIO, bytes, str]=None,
                is_approx: bool=True,
                deleter: str=None) -> None:
        """Delete entities by Primary key ID, or (for exact:) by file or URI

        Args:
            collection_name (str): The collection name
            ids (Union[List[str], List[int]], optional): The list of IDs for which is to happen
            urls (List[str], optional): A list of URLs to be hashed & potentially deleted. Defaults to None.
            file (Union[BinaryIO, bytes, str], optional): File to be hashed & potentially deleted. Defaults to None.
            is_approx (bool): Is the collection approximate or exact?

        Returns:
            _type_: None
        """
        if is_approx:
            raise NotImplementedError("Soft-delete entries has not been implemented for approx yet")
        else:
            # Exact
            if ids:
                return soft_delete_entries_exact_entries_soft_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    body=BodySoftDeleteEntriesExactEntriesSoftDeleteDelete(entry_ids=ids),
                    deleter=deleter
                )
            if urls:
                return soft_delete_entries_exact_entries_soft_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    body=BodySoftDeleteEntriesExactEntriesSoftDeleteDelete(urls=urls),
                    deleter=deleter
                )
            if file:
                return soft_delete_entries_by_file_exact_entries_by_file_soft_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    multipart_data=BodySoftDeleteEntriesByFileExactEntriesByFileSoftDeleteDelete(
                        file=fastAPIFile(payload=get_tmp_file(file), file_name=str(uuid4()),
                                                mime_type="image/jpg")
                    ),
                    deleter=deleter
                )

    @propagate_exception
    def get_trashcan_entries(self,
                             deleter: str=None,
                             exact: bool=True,
                             start_date: date=None,
                             end_date: date=None,
                             order_by: str=None,
                             desc: bool=True,
                             limit: int=1000):
        if not exact:
            raise NotImplementedError("Getting trash can entries is not supported for approx.")
        else:
            return trashcan_entries_exact_trashcan_entries_get.sync_detailed(client=self.client,
                                                                             deleter=deleter,
                                                                             start_date=start_date,
                                                                             end_date=end_date,
                                                                             order_by=order_by,
                                                                             desc=desc,
                                                                             limit=limit)
        
    @propagate_exception
    def hard_delete_entities(self, collection_name: str,
                ids: Union[List[str], List[int]]=None,
                urls: List[str]=None,
                file: Union[BinaryIO, bytes, str]=None,
                is_approx: bool=True) -> None:
        """Delete entities by Primary key ID, or (for exact:) by file or URI

        Args:
            collection_name (str): The collection name
            ids (Union[List[str], List[int]], optional): The list of IDs for which is to happen
            urls (List[str], optional): A list of URLs to be hashed & potentially deleted. Defaults to None.
            file (Union[BinaryIO, bytes, str], optional): File to be hashed & potentially deleted. Defaults to None.
            is_approx (bool): Is the collection approximate or exact?

        Returns:
            _type_: None
        """
        if is_approx:
            return delete_entities_collection_entities_delete.sync_detailed(
                client=self.client, name=collection_name, body=ids)
        else:
            # Exact
            if ids:
                return hard_delete_entries_exact_entries_hard_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    body=BodyHardDeleteEntriesExactEntriesHardDeleteDelete(entry_ids=ids)
                )
            if urls:
                return hard_delete_entries_exact_entries_hard_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    body=BodyHardDeleteEntriesExactEntriesHardDeleteDelete(urls=urls),
                )
            if file:
                return hard_delete_entries_by_file_exact_entries_by_file_hard_delete_delete.sync_detailed(
                    client=self.client, collection_name=collection_name,
                    multipart_data=BodyHardDeleteEntriesByFileExactEntriesByFileHardDeleteDelete(
                        file=fastAPIFile(payload=get_tmp_file(file), file_name=str(uuid4()),
                                                mime_type="image/jpg")
                    )
                )

    @propagate_exception
    def sample_entities(self, collection_name: str, max_samples: int = 20):
        """Sample entities of a collection

        Args:
            collection_name (str): The collection name
            max_samples (int, optional): The maximum number of entities to sample. Defaults to 20.

        Returns:
            _type_: _description_
        """

        return sample_collection_sample_get.sync_detailed(client=self.client,
                collection_name=collection_name, max_samples=max_samples)


    @propagate_exception
    def sizeof(self, name: str, exact: bool=False) -> int:
        """Get the size of (number of entities) in a collection

        Args:
            name (str): The collection name
            exact (bool): is the collection an exact lookup?

        Raises:
            RuntimeError: If collection does not exist

        Returns:
            int: The size of the collection
        """
        return sizeof_collection_collection_size_name_get.sync_detailed(client=self.client, name=name, exact=exact)


    @propagate_exception
    def ingest(self, data: InputData) -> None:
        """Ingest data into a collection

        Args:
            data (InputData): The data to be ingested

        Returns:
            _type_: None
        """
        return ingest_ingest_post.sync_detailed(client=self.client, body=data)


    @propagate_exception
    def ingest_with_file(self, collection_name: str, json_input_data: str,
                    file: Union[str, bytes, BinaryIO]=None) -> None:
        """Ingest *a single* element with a file into a collection

        Args:
            collection_name (str): The unique name of the collection
            json_input_data (str): A json string formatted input identical to that of InputData
            file (Union[str, bytes, BinaryIO], optional): _description_. Defaults to None.

        Returns:
            _type_: None
        """

        file_name: str = "tmpfile"
        if isinstance(file, str):
            try:
                file_name = os.path.basename(file)
            except:
                pass # we did our best \_(ツ)_/¯

        return ingest_w_file_ingest_file_post.sync_detailed(client=self.client, multipart_data=BodyIngestWFileIngestFilePost(
            file=fastAPIFile(payload=get_tmp_file(file), file_name=file_name, mime_type="image/jpg")),
            collection_name=collection_name, json_input_data=json_input_data
        )

    @propagate_exception
    def create_index(self, index_spec: IndexInSpec) -> None:
        """Create an index for one field within the collection

        Args:
            index_spec (IndexInSpec): The creation spec of the index

        Returns:
            _type_: None
        """
        return create_index_create_post.sync_detailed(client=self.client, body=index_spec)


    @propagate_exception
    def drop_index(self, collection_name: str, field_name: str) -> None:
        """Drop (delete) a previously created index

        Args:
            collection_name (str): The collection name
            field_name (str): The field name on which the index is defined

        Returns:
            _type_: None
        """
        return delete_index_collection_name_field_name_delete.sync_detailed(client=self.client, collection_name=collection_name, field_name=field_name)

    @propagate_exception
    def query(self, query_spec: Union[QuerySpec, VectorQuerySpec]) -> Union[List[Dict], QueryResponse]:
        """Query a collection using either a vector, a batch of vectors or query on the fields

        Args:
            query_spec (Union[QuerySpec, VectorQuerySpec]): The specification of the query

        Raises:
            RuntimeError: If the query spec is non conformant, this is raised.

        Returns:
            Union[List[Dict], QueryResponse]: Either a list with the queried fields or a QueryResponse object from a vector query
        """

        if isinstance(query_spec, QuerySpec):
            return query_query_fields_post.sync_detailed(client=self.client, body=query_spec)
        elif isinstance(query_spec, VectorQuerySpec):
            return query_query_vectors_post.sync_detailed(client=self.client, body=query_spec)
        else:
            raise RuntimeError(f"Unable to query with given query spec of type {type(query_spec)}")

    @propagate_exception
    def recover_collection(self, collection_name: str, is_approx: bool = False) -> None:
        if is_approx:
            raise NotImplementedError("Approximate collections not yet supported")
        return restore_collection_exact_collection_restore_patch.sync_detailed(client=self.client, name=collection_name)

    @propagate_exception
    def restore_entries(self, collection_name: str,
                ids: Union[List[str], List[int]]=None,
                is_approx: bool=True) -> None:
        if is_approx:
            raise NotImplementedError("Approximate collections not yet supported")
        if ids:
            return restore_entries_exact_entries_restore_patch.sync_detailed(
                client=self.client, collection_name=collection_name,
                body=ids
            )

    @propagate_exception
    def put_exact(self, collection_name: str,
                    foreign_key: Union[str, int]=None,
                    file: Union[str, bytes, BinaryIO]=None,
                    uri: str=None,
                    metadata: dict=None,
                    upsert: bool = False,
                    timestamp: str = None,
                    store_raw_data: bool = False,
                    ingestor: str=None
                    ) -> str:
        """Put an entry into an exact store collection

        Args:
            collection_name (str): The collection name
            foreign_key (Union[str, int], optional): Foreign. key Defaults to None.
            file (Union[str, bytes, BinaryIO], optional): The file containing the data. Defaults to None.
            uri (str, optional): The URI of input data. Defaults to None.
            metadata (dict, optional): Dictionary of additional metadata to store with entry. Defaults to None.
            upsert (bool, optional): Update entry if it already exists. Defaults to None.
            timestamp (str, optional): Timestamp to associate with entry. Defaults to None.
            store_raw_data(bool, optional): Store the raw data directly in lectric. Defaults to False.

        Raises:
            RuntimeError: When input file or URI is unprocessable

        Returns:
            str: The inserted hash
        """
        if uri:
            return put_entry_by_uri_exact_hash_collection_name_uri_put.sync_detailed(
                collection_name=collection_name, client=self.client,
                foreign_key=foreign_key, uri=uri,
                upsert=upsert,
                metadata=None if not metadata else json.dumps(metadata),
                timestamp=timestamp,
                store_raw_data=store_raw_data,
                ingestor=ingestor)
        elif file:
            return put_entry_by_file_exact_hash_collection_name_file_put.sync_detailed(
                collection_name=collection_name, client=self.client,
                multipart_data=BodyPutEntryByFileExactHashCollectionNameFilePut(
                        file=fastAPIFile(payload=get_tmp_file(file), file_name="tmpfile", mime_type="image/jpg")
                    ), foreign_key = foreign_key, metadata=None if not metadata else json.dumps(metadata),
                        upsert=upsert, timestamp=timestamp,
                        store_raw_data=store_raw_data,
                        ingestor=ingestor
                )

    @propagate_exception
    def lookup_exact(self, collection_name: str,
                    file: Union[str, bytes, BinaryIO]=None,
                    uri: str=None,
                    fk: str=None) -> Union[List[str], None]:
        """Lookup exact entry in a collection

        Args:
            collection_name (str): The name of the collection
            file (Union[str, bytes, BinaryIO], optional): Input file data. Defaults to None.
            uri (str, optional): Input URI data. Defaults to None.

        Raises:
            RuntimeError: When input file or URI is unprocessable

        Returns:
            Union[List[str], None]: The found match or None
        """
        if uri:
            return lookup_by_uri_exact_lookup_hash_uri_collection_name_get.sync_detailed(
                collection_name=collection_name, client=self.client, uri=uri, fk=fk,
            )
        elif file:
            return lookup_by_file_exact_lookup_file_collection_name_post.sync_detailed(
                collection_name=collection_name, client=self.client,
                multipart_data=BodyLookupByFileExactLookupFileCollectionNamePost(
                        file=fastAPIFile(payload=get_tmp_file(file), file_name="tmpfile", mime_type="image/jpg")
                    )
                )

    @propagate_exception
    def get_exact_hash_algos(self) -> List[str]:
        """Get a list of the available hashing algorithms

        Returns:
            List[str]: A list of the hash algos
        """
        return get_available_algos_exact_algos_get.sync_detailed(client=self.client)

    @propagate_exception
    def is_exact_algos_available(self, algo: str) -> bool:
        """Determine whether a specific hashing algorithm is supported

        Args:
            algo (str): The algorithm name

        Returns:
            bool: True if the algorithm is supported/available for use
        """
        return is_algo_available_exact_available_algo_get.sync_detailed(client=self.client)

    @propagate_exception
    def get_exact_collection_entries(self,
                    collection_name: str,
                    ingestor: str = None,
                    start_date: date = None,
                    end_date: date = None,
                    ingest_source: str = None,
                    traits: str = None,
                    order_by: str = None,
                    desc: bool = True,
                    limit: int = 1000) -> List[ExactCollectionEntry]:
        """
        Get exact entries in a collection_name

        Args:
            ingestor (str): Email address of the ingestor
            collection_name (str): Collection Name
            start_date (date): Start date (UTC)
            end_date (date): End date (UTC)
            order_by (str): Order the results by this field. Defaults to None.
            desc (bool): Order descending? or ascending. Defaults to True.
            limit (int): the max results to return. Defaults to 1000.

        Returns:
            List[Dict]: entries in a given collection_name collection, ingested by ingestor (if provided), contained within the date ranges (if provided).
        """
        return get_exact_collection_entries_exact_collection_entries_get.sync_detailed(client=self.client,
                                                                                       collection_name=collection_name,
                                                                                       ingestor=ingestor,
                                                                                       start_date=start_date,
                                                                                       end_date=end_date,
                                                                                       ingest_source=ingest_source,
                                                                                       traits=traits,
                                                                                       order_by=order_by,
                                                                                       desc=desc,
                                                                                       limit=limit)

    @propagate_exception
    def exact_select(self, collection_name: str,
                        what: str = "*",
                        where: str = None,
                        order_by: str = None,
                        desc: bool = True,
                        limit: int = 10000) -> List[Dict]:
        """Select from collection using SQL logic

        Args:
            collection_name (str): The unique name of the collection
            what (str, optional): what columns/fields to select. Defaults to "*".
            where (str, optional): the where predicate. Defaults to None.
            order_by (str, optional): order the results by this field. Defaults to None.
            desc (bool, optional): order descending? or ascending. Defaults to True.
            limit (int, optional): the max results to return. Defaults to 10000.

        Returns:
            List[Dict]: entries matching the query from the database
        """

        return select_entries_exact_select_post.sync_detailed(
            client=self.client, collection_name=collection_name,
            what=what, where=where, order_by=order_by, desc=desc, limit=limit
        )

    def _pformat_collection(self, collection: Collection) -> str:
        return f"""
                Collection name: {collection.name}
                Size (estimated): {self.sizeof(collection.name)}
                Index(es): {self.get_indexes(collection.name)}
                {pformat_collection_schema(collection.coll_schema)}
                """

    @propagate_exception
    def _exact_info(self, collection_name: str):
         return info_exact_info_get.sync_detailed(client=self.client, name=collection_name)


    def info(self, collection_name: str, exact: bool = False) -> str:
        """Get info on a collection

        Args:
            collection_name (str): The unique name of the collection
            exact (bool, optional): Exact collections only? Defaults to False.
        """
        if exact:
            return self._exact_info(collection_name)
        else:
            return self._pformat_collection(self.get_collection(collection_name))

    @propagate_exception
    def has_reverse_lookup(self, collection_name: str) -> bool:
        """Does the exact collection have reverse lookup enabled

        Args:
            collection_name (str): The unique name of the exact collection

        Returns:
            bool: True if so, else False
        """
        return has_revlookup_exact_revlookup_get.sync_detailed(client=self.client, name=collection_name)


    @propagate_exception
    def reverse_lookup(self, collection_name: str, limit: int = None) -> List[Tuple]:
        """Get elements from the reverse lookup

        Args:
            collection_name (str): The unique name of the collection
            limit (int, optional): Limit the number of outputs. Defaults to None.

        Returns:
            List[Tuple]: The returned elements in the reverse lookup
        """

        return revlookup_list_exact_revlookup_list_get.sync_detailed(
                        client=self.client, name=collection_name, limit=limit)

    @propagate_exception
    def has_dupe_counter(self, collection_name: str) -> bool:
        """Does the exact collection have duplicate counting enabled

        Args:
            collection_name (str): The unique name of the exact collection

        Returns:
            bool: True if so, else False
        """
        return has_dupecounter_exact_dupecounter_get.sync_detailed(client=self.client, name=collection_name)


    @propagate_exception
    def get_dupes(self, collection_name: str, limit: int = None, total_only: bool=False) -> Union[List[Tuple], int]:
        """Get duplicates from exact collections with counter enabled

        Args:
            collection_name (str): The unique name of the collection
            limit (int, optional): The maximum number of entries to return. Defaults to None.
            total_only (bool, optional): Provide only the total count of duplicates. Defaults to False.

        Returns:
            Union[List[Tuple], int]: Either the duplicates requested or the total only
        """

        return dupes_list_exact_counter_dupes_get.sync_detailed(
                                client=self.client, name=collection_name, limit=limit, total_only=total_only)

    @propagate_exception
    def raw_query(self, query: str) -> List[Tuple]:
        """EXPERIMENTAL: Query directly into the underlying storage for data. Use SQL syntax

        Args:
            query (str): The query string

        Returns:
            List[Tuple]: Database rows response
        """
        return raw_query_exact_experimental_raw_query_post.sync_detailed(client=self.client, query=query)

    @propagate_exception
    def has_exact_engine(self) -> bool:
        """Check if the client has an exact engine active

        Returns:
            bool: If the exact engine is configured and available
        """
        return  ping_exact_ping_post.sync_detailed(client=self.client)

    @propagate_exception # TODO: Hack! We should use get_collection and populate accordingly
    def get_schema_info(self, collection_name: str) -> SchemaInfoExactSchemaInfoGetResponseSchemaInfoExactSchemaInfoGet:
        """Get the underlying storage schema

        Args:
            collection_name (str): The name of the collection

        Returns:
            SchemaInfoExactSchemaInfoGetResponseSchemaInfoExactSchemaInfoGet: A representation of each column in the schema with its type
        """
        return schema_info_exact_schema_info_get.sync_detailed(client=self.client, name=collection_name)

    @propagate_exception
    def get_url(self, path: str, access_duration_sec: float=60*60) -> str:
        """Get the URL of data stored within Lectric given a file path. Obtain the file path through a query on an collection that has a link field schema

        Args:
            path (str): The file path within the collection
            access_duration_sec (float): How long the url will be active
        Returns:
            str: a fully defined url
        """

        return get_url_collection_get_url_get.sync_detailed(client=self.client,
                                                    path=path, access_duration_sec=access_duration_sec)

    @propagate_exception
    def get_collection_metadata(self, collection_name: str, is_approx: bool=False) -> CollectionMetadata:
        """Retrieves the collection metadata, which includes the ETag stored in the form of the last updated timestamp.

        Args:
            collection_name (str): Collection Name
            is_approx (bool): Is the collection an approximate one or not
        Returns:
            str: Collection Metadata
        """
        if is_approx:
            raise NotImplementedError("Approximate collections not yet supported")
        else:
            return get_collection_metadata_exact_collection_metadata_get.sync_detailed(
                                        client=self.client, collection_name=collection_name)


from typing import Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="InputData")


@_attrs_define
class InputData:
    """
    Attributes:
        collection_name (str):
        data (List[List[Any]]):
        object_type (Union[Literal['InputData'], Unset]):  Default: 'InputData'.
    """

    collection_name: str
    data: List[List[Any]]
    object_type: Union[Literal["InputData"], Unset] = "InputData"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        collection_name = self.collection_name

        data = []
        for data_item_data in self.data:
            data_item = data_item_data

            data.append(data_item)

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "collection_name": collection_name,
                "data": data,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        collection_name = d.pop("collection_name")

        data = []
        _data = d.pop("data")
        for data_item_data in _data:
            data_item = cast(List[Any], data_item_data)

            data.append(data_item)

        object_type = d.pop("object_type", UNSET)

        input_data = cls(
            collection_name=collection_name,
            data=data,
            object_type=object_type,
        )

        input_data.additional_properties = d
        return input_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import List
import pytest
from .utils import (
    del_collection_if_exists,
    gen_random_fp_collection_spec,
    gen_random_name,
    gen_random_fp_vectors,
    create_collection)
import lectric as lc

COLL_DIM = 12

@pytest.fixture(scope='module')
def collection_name() -> str:
    return gen_random_name(10)

@pytest.fixture(scope='module')
def collection_spec() -> lc.CollectionInSpec:
    spec = gen_random_fp_collection_spec(COLL_DIM)
    yield spec

@pytest.fixture(scope='module')
def collection(client: lc.LectricClient, collection_spec: lc.CollectionInSpec) -> lc.Collection:
    coll: lc.Collection = create_collection(client, collection_spec)

    yield coll
    del_collection_if_exists(coll.name, client)

def test_index(client: lc.LectricClient, collection: lc.Collection):

    assert client.is_collection_empty(collection.name)

    num_elements = 200
    vecs: List[List[float]] = gen_random_fp_vectors(COLL_DIM, num_elements)

    input_data: lc.InputData = lc.InputData(
        collection_name=collection.name,
        data=[vecs]
    )

    client.ingest(input_data)
    # Get the collection and test it's not empty
    assert not client.is_collection_empty(collection.name)

    vec_field: str = "vecs"

    index_spec: lc.IndexInSpec = lc.IndexInSpec(
        collection_name=collection.name,
        field_name=vec_field,
        index=lc.Index(index_type=lc.IndexType.IVF_FLAT,
                    metric_type=lc.VectorSpace.L2,
                    params=lc.IndexParams())
    )

    # Test create index
    client.create_index(index_spec)


    # Test get indexes defined for a collection
    indexes: List[str] = client.get_indexes(collection.name)

    assert len(indexes) == 1
    assert indexes[0] == index_spec.field_name


    # Query index ...
    k = 5

    res: lc.QueryResponse = client.query(lc.VectorQuerySpec(
        data=vecs[:2],
        collection_name=collection.name,
        search_field=vec_field,
        search_params=lc.QueryParams(metric_type=lc.VectorSpace.L2,
                                params=lc.QueryMetaParams(
                                nprobe=2
                            )),
        output_fields=[],
        limit=k)

        )
    assert len(res.hits) == 2
    assert len(res.hits[0]) == k

    # Test deleting the index
    client.drop_index(collection.name, index_spec.field_name)

    # Test deletion is successful
    assert len(client.get_indexes(collection.name)) == 0



import pytest
import lectric as lc
from typing import List
from .utils import (
    del_collection_if_exists,
)
import uuid

PRIMARY_FIELD_NAME = "id"
FOREIGN_ID_FIELD_NAME = "fk"
LINK_FIELD_NAME = "link"
METADATA_FIELD_NAME = "metadata"
IS_FILE_UPLOAD_FIELD_NAME = "is_file_upload"
IS_MARKED_FOR_DELETION = "is_marked_for_deletion"
INGESTOR = "ingestor"
SOFT_DELETER = "soft_deleter"
TEST_COLLECTION_NAME = 'test'
USERNAME = 'dummy@dummy.com'

TEST_IMAGE_URI = 'https://picsum.photos/seed/picsum/200/300'
HASH_ALGO= 'md5'
TEST_METADATA = {'traits': 'Artistic Imagery'}
SOFT_DELETED_ENTRIES_SIZE = 'Soft-deleted entries size'

@pytest.fixture(scope='function')
def collection(client: lc.LectricClient) -> lc.Collection:
    fields: List[lc.FieldSchema] = [lc.FieldSchema(PRIMARY_FIELD_NAME, lc.DataType.VARCHAR,
                                                                                max_length=128, is_primary=True)]
    fields.append(lc.FieldSchema(name=LINK_FIELD_NAME, dtype=lc.DataType.STRING, description="link field"))
    fields.append(lc.FieldSchema(name=METADATA_FIELD_NAME, dtype=lc.DataType.JSON, description="JSONB metadata"))
    fields.append(lc.FieldSchema(name=IS_FILE_UPLOAD_FIELD_NAME, dtype=lc.DataType.BOOL, description="Is File Upload"))
    fields.append(lc.FieldSchema(name=IS_MARKED_FOR_DELETION, dtype=lc.DataType.BOOL, description="Is marked for deletion"))
    fields.append(lc.FieldSchema(name=INGESTOR, dtype=lc.DataType.STRING, description="MSFT alias of the ingestor"))
    fields.append(lc.FieldSchema(name=FOREIGN_ID_FIELD_NAME, dtype=lc.DataType.STRING, description="Foreign key"))

    enable_dupe_counter: bool = False
    collection_name = 'test__' + str(uuid.uuid4().hex[0:6])
    collection = client.create_collection(lc.CollectionInSpec(name=collection_name, approx=False,
                                            coll_schema=lc.CollectionSchema(fields=fields)),
                                            with_dupe_counter=enable_dupe_counter,
                                            hash_algo=lc.HashAlgo[HASH_ALGO])
    yield collection

    del_collection_if_exists(collection.name, client, exact=True)

@pytest.fixture(scope='function')
def exact_entry_id(collection: lc.Collection, client: lc.LectricClient) -> str:
    entry_id = client.put_exact(collection_name=collection.name,
                    uri=TEST_IMAGE_URI,
                    metadata=TEST_METADATA,
                    upsert=True,
                    store_raw_data=True,
                    ingestor=USERNAME)
    yield entry_id

    client.hard_delete_entities(collection_name=collection.name, ids=[entry_id], is_approx=False)

def test_ingest(client: lc.LectricClient, exact_entry_id: str, collection: lc.Collection):
    lookup_result = client.lookup_exact(collection.name, uri=TEST_IMAGE_URI)

    assert lookup_result[METADATA_FIELD_NAME] == TEST_METADATA
    assert lookup_result[PRIMARY_FIELD_NAME] == exact_entry_id
    assert lookup_result[INGESTOR] == USERNAME


# check the collection info after doing soft_delete
def test_soft_delete(client: lc.LectricClient, exact_entry_id: str, collection: lc.Collection):
    client.soft_delete_entities(collection_name=collection.name, ids=[exact_entry_id], is_approx=False, deleter=USERNAME)
    collection_info_str = client.info(collection_name=collection.name, exact=True)
    collection_info_dict = convert_info_str_to_dict(collection_info_str=collection_info_str)
    lookup_result = client.lookup_exact(collection.name, uri=TEST_IMAGE_URI)
    trashcan_result = client.get_trashcan_entries(deleter=USERNAME)

    assert int(collection_info_dict[SOFT_DELETED_ENTRIES_SIZE]) == 1
    assert lookup_result[IS_MARKED_FOR_DELETION] == True
    assert lookup_result[INGESTOR] == USERNAME

    assert len(trashcan_result) == 1
    assert trashcan_result[0].collection == collection.name
    assert trashcan_result[0].data_collection_primary_key == exact_entry_id
    assert trashcan_result[0].soft_deleter == USERNAME

def test_soft_delete_then_recover(client: lc.LectricClient, exact_entry_id: str, collection: lc.Collection):
    lookup_result = client.lookup_exact(collection.name, uri=TEST_IMAGE_URI)
    assert lookup_result[PRIMARY_FIELD_NAME] == exact_entry_id
    assert lookup_result[INGESTOR] == USERNAME

    # soft delete
    client.soft_delete_entities(collection_name=collection.name, ids=[exact_entry_id], is_approx=False, deleter=USERNAME)
    lookup_result = client.lookup_exact(collection.name, uri=TEST_IMAGE_URI)
    collection_info_str = client.info(collection_name=collection.name, exact=True)
    collection_info_dict = convert_info_str_to_dict(collection_info_str=collection_info_str)
    trashcan_result = client.get_trashcan_entries(deleter=USERNAME)

    assert len(trashcan_result) == 1
    assert trashcan_result[0].collection == collection.name
    assert trashcan_result[0].data_collection_primary_key == exact_entry_id
    assert trashcan_result[0].soft_deleter == USERNAME

    assert int(collection_info_dict[SOFT_DELETED_ENTRIES_SIZE]) == 1
    assert lookup_result[IS_MARKED_FOR_DELETION] == True
    assert lookup_result[INGESTOR] == USERNAME

    # recover the entry that's deleted
    client.restore_entries(collection_name=collection.name, ids=[exact_entry_id], is_approx=False)
    lookup_result = client.lookup_exact(collection.name, uri=TEST_IMAGE_URI)
    collection_info_str = client.info(collection_name=collection.name, exact=True)
    collection_info_dict = convert_info_str_to_dict(collection_info_str=collection_info_str)
    trashcan_result = client.get_trashcan_entries(deleter=USERNAME)

    assert len(trashcan_result) == 0
    assert int(collection_info_dict[SOFT_DELETED_ENTRIES_SIZE]) == 0
    assert lookup_result[IS_MARKED_FOR_DELETION] == False
    assert lookup_result[INGESTOR] == USERNAME

def convert_info_str_to_dict(collection_info_str: str) -> dict:
    collection_info_dict = {}
    for item in collection_info_str:
        # Split each item into key and value
        key, value = item.split(': ', 1)
        # Remove any leading or trailing whitespace from the value
        value = value.strip()
        # Add the key-value pair to the dictionary
        collection_info_dict[key] = value
    return collection_info_dict



import pytest
from typing import List

from .utils import *
import lectric as lc

COLL_DIM = 8

@pytest.fixture(scope='module')
def collection_name() -> str:
    coll_name: str = gen_random_name(10)
    return coll_name


@pytest.fixture(scope="module")
def collection_in_spec(collection_name: str) -> lc.CollectionInSpec:
    fields: List[lc.FieldSchema] = [
        lc.FieldSchema(name="id", dtype=lc.DataType.INT64, is_primary=True,
            auto_id=True, description="Primary key"),
        lc.FieldSchema(name="vecs", dtype=lc.DataType.FLOAT_VECTOR,
            dim=COLL_DIM, description="Data embeddings")
    ]
    return lc.CollectionInSpec(collection_name,
            coll_schema=lc.CollectionSchema(fields=fields, description="Test Schema description"))

# Also a test for create
@pytest.fixture(scope="module")
def collection(client: lc.LectricClient, collection_in_spec: lc.CollectionInSpec) -> lc.Collection:
    del_collection_if_exists(collection_in_spec.name, client)

    # Create
    coll: lc.Collection = client.create_collection(collection_in_spec)

    assert coll.name == collection_in_spec.name
    assert coll.coll_schema.to_dict() == collection_in_spec.coll_schema.to_dict()
    yield coll

    del_collection_if_exists(collection_in_spec.name, client)


@pytest.fixture(scope="module")
def three_collections(client: lc.LectricClient) -> List[lc.Collection]:
    NUM_COLLS = 3

    coll_in_specs = [gen_random_binary_collection_spec(COLL_DIM) for _ in range(NUM_COLLS)]

    collections: List[lc.Collection] = []

    for coll_in_spec in coll_in_specs:
        coll: lc.Collection = client.create_collection(coll_in_spec)

        assert coll.name == coll_in_spec.name
        assert coll.consistency_level == coll_in_spec.consistency_level
        assert coll.coll_schema.to_dict() == coll_in_spec.coll_schema.to_dict()

        collections.append(coll)

    yield collections

    # Delete the ones just created them
    for coll in collections:
        del_collection_if_exists(coll.name, client)


# NOTE: Done in fixtures
def test_create(collection: lc.Collection):
    assert collection

def test_empty(client: lc.LectricClient , collection: lc.Collection):
    assert client.is_collection_empty(collection.name)

def test_size(client: lc.LectricClient , collection: lc.Collection):
    assert client.sizeof(collection.name) == 0

def test_get(client: lc.LectricClient, collection: lc.Collection):
    coll: lc.Collection = client.get_collection(collection.name)
    assert coll.name == collection.name


def test_list(client: lc.LectricClient, three_collections: List[lc.Collection]):
    q_colls: List[lc.Collection] = client.list_collections()
    assert len(q_colls) >= len(three_collections)


# NOTE: Done in fixtures
def test_delete(collection: lc.Collection):
    assert collection




import lectric as lc
import pytest
from typing import List

from .utils import (
    del_collection_if_exists,
    gen_random_fp_collection_spec,
    gen_random_name,
    gen_random_fp_vectors,
    create_collection)

COLL_DIM = 12

@pytest.fixture(scope='module')
def collection_name() -> str:
    return gen_random_name(10)

@pytest.fixture(scope='module')
def collection_spec() -> lc.CollectionInSpec:
    spec = gen_random_fp_collection_spec(COLL_DIM)
    yield spec

@pytest.fixture(scope='module')
def collection(client: lc.LectricClient, collection_spec: lc.CollectionInSpec) -> lc.Collection:
    coll: lc.Collection = create_collection(client, collection_spec)

    yield coll
    del_collection_if_exists(coll.name, client)


def test_ingest(client: lc.LectricClient, collection: lc.Collection):

    num_elements = 200
    vecs: List[List[float]] = gen_random_fp_vectors(COLL_DIM, num_elements)

    input_data: lc.InputData = lc.InputData(
        collection_name=collection.name,
        data=[vecs]
    )

    # Test the collection is empty
    assert client.is_collection_empty(collection.name)

    # Ingest
    client.ingest(input_data)

    # Get the collection and test it's not empty
    assert not client.is_collection_empty(collection.name)


@pytest.mark.parametrize("pk_dtype",
            [
                lc.DataType.INT64,
                lc.DataType.VARCHAR
            ]
            )
def test_delete_collection_entity(client: lc.LectricClient, pk_dtype: lc.DataType):
    coll_dim = 8
    num_entities = 5
    num_del_entities = 2

    coll: lc.Collection = None

    # NOTE: Consistency level at Strong is necessary for deletion
    spec: lc.CollectionInSpec = gen_random_fp_collection_spec(coll_dim=coll_dim,
                            random_field=False, auto_id=False, consistency_level="Strong",
                            pk_type=pk_dtype)
    try:
        ids = []
        coll = client.create_collection(spec)

        assert client.is_collection_empty(coll.name)

        if pk_dtype == lc.DataType.INT64:
            ids = list(range(num_entities))

        elif pk_dtype == lc.DataType.VARCHAR:
            # NOTE: If you use IDs that can be cast to an int they might be turned into ints within the vdb
            ids = [ gen_random_name(4) for _ in  range(num_entities) ]

        input_data = lc.InputData(
            collection_name=coll.name,
            data=[ids, gen_random_fp_vectors(coll_dim, num_entities)]
        )

        assert input_data
        client.ingest(input_data)

        assert client.sizeof(coll.name) == num_entities

        client.hard_delete_entities(coll.name, ids=ids[:num_del_entities])
        # assert client.sizeof(coll.name) == (num_entities-num_del_entities) # FIXME
    finally:
        if coll:
            client.hard_drop_collection(coll.name)




from typing import Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="IndexParams")


@_attrs_define
class IndexParams:
    """
    Attributes:
        object_type (Union[Literal['IndexParams'], Unset]):  Default: 'IndexParams'.
        nlist (Union[None, Unset, int]): Number of cluster units Default: 128.
        quant (Union[None, Unset, int]): Number of factors of product quantization
        nbits (Union[None, Unset, int]): Number of bits in which each low-dimensional vector is stored
        m (Union[None, Unset, int]): Maximum degree of the node
        ef_construction (Union[None, Unset, int]): Search scope
        pqm (Union[None, Unset, int]): Number of factors of product quantization
        ntrees (Union[None, Unset, int]): The number of methods of space division
    """

    object_type: Union[Literal["IndexParams"], Unset] = "IndexParams"
    nlist: Union[None, Unset, int] = 128
    quant: Union[None, Unset, int] = UNSET
    nbits: Union[None, Unset, int] = UNSET
    m: Union[None, Unset, int] = UNSET
    ef_construction: Union[None, Unset, int] = UNSET
    pqm: Union[None, Unset, int] = UNSET
    ntrees: Union[None, Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        object_type = self.object_type

        nlist: Union[None, Unset, int]
        if isinstance(self.nlist, Unset):
            nlist = UNSET
        else:
            nlist = self.nlist

        quant: Union[None, Unset, int]
        if isinstance(self.quant, Unset):
            quant = UNSET
        else:
            quant = self.quant

        nbits: Union[None, Unset, int]
        if isinstance(self.nbits, Unset):
            nbits = UNSET
        else:
            nbits = self.nbits

        m: Union[None, Unset, int]
        if isinstance(self.m, Unset):
            m = UNSET
        else:
            m = self.m

        ef_construction: Union[None, Unset, int]
        if isinstance(self.ef_construction, Unset):
            ef_construction = UNSET
        else:
            ef_construction = self.ef_construction

        pqm: Union[None, Unset, int]
        if isinstance(self.pqm, Unset):
            pqm = UNSET
        else:
            pqm = self.pqm

        ntrees: Union[None, Unset, int]
        if isinstance(self.ntrees, Unset):
            ntrees = UNSET
        else:
            ntrees = self.ntrees

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if nlist is not UNSET:
            field_dict["nlist"] = nlist
        if quant is not UNSET:
            field_dict["quant"] = quant
        if nbits is not UNSET:
            field_dict["nbits"] = nbits
        if m is not UNSET:
            field_dict["M"] = m
        if ef_construction is not UNSET:
            field_dict["efConstruction"] = ef_construction
        if pqm is not UNSET:
            field_dict["PQM"] = pqm
        if ntrees is not UNSET:
            field_dict["ntrees"] = ntrees

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        object_type = d.pop("object_type", UNSET)

        def _parse_nlist(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        nlist = _parse_nlist(d.pop("nlist", UNSET))

        def _parse_quant(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        quant = _parse_quant(d.pop("quant", UNSET))

        def _parse_nbits(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        nbits = _parse_nbits(d.pop("nbits", UNSET))

        def _parse_m(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        m = _parse_m(d.pop("M", UNSET))

        def _parse_ef_construction(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        ef_construction = _parse_ef_construction(d.pop("efConstruction", UNSET))

        def _parse_pqm(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        pqm = _parse_pqm(d.pop("PQM", UNSET))

        def _parse_ntrees(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        ntrees = _parse_ntrees(d.pop("ntrees", UNSET))

        index_params = cls(
            object_type=object_type,
            nlist=nlist,
            quant=quant,
            nbits=nbits,
            m=m,
            ef_construction=ef_construction,
            pqm=pqm,
            ntrees=ntrees,
        )

        index_params.additional_properties = d
        return index_params

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="BodyHardDeleteEntriesExactEntriesHardDeleteDelete")


@_attrs_define
class BodyHardDeleteEntriesExactEntriesHardDeleteDelete:
    """
    Attributes:
        entry_ids (Union[List[int], List[str], Unset]):
        urls (Union[Unset, List[str]]):
    """

    entry_ids: Union[List[int], List[str], Unset] = UNSET
    urls: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        entry_ids: Union[List[int], List[str], Unset]
        if isinstance(self.entry_ids, Unset):
            entry_ids = UNSET
        elif isinstance(self.entry_ids, list):
            entry_ids = self.entry_ids

        else:
            entry_ids = self.entry_ids

        urls: Union[Unset, List[str]] = UNSET
        if not isinstance(self.urls, Unset):
            urls = self.urls

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if entry_ids is not UNSET:
            field_dict["entry_ids"] = entry_ids
        if urls is not UNSET:
            field_dict["urls"] = urls

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_entry_ids(data: object) -> Union[List[int], List[str], Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                entry_ids_type_0 = cast(List[int], data)

                return entry_ids_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, list):
                raise TypeError()
            entry_ids_type_1 = cast(List[str], data)

            return entry_ids_type_1

        entry_ids = _parse_entry_ids(d.pop("entry_ids", UNSET))

        urls = cast(List[str], d.pop("urls", UNSET))

        body_hard_delete_entries_exact_entries_hard_delete_delete = cls(
            entry_ids=entry_ids,
            urls=urls,
        )

        body_hard_delete_entries_exact_entries_hard_delete_delete.additional_properties = d
        return body_hard_delete_entries_exact_entries_hard_delete_delete

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# TODO: Eliminate dupe
from enum import Enum, IntEnum, unique, auto

class DataType(IntEnum):
    NONE = 0
    BOOL = 1
    INT8 = 2
    INT16 = 3
    INT32 = 4
    INT64 = 5

    FLOAT = 10
    DOUBLE = 11

    STRING = 20
    VARCHAR = 21
    JSON = 23
    BINARY_VECTOR = 100
    FLOAT_VECTOR = 101

    UNKNOWN = 999

@unique
class IndexType(str, Enum):
    FLAT = "FLAT"
    IVF_FLAT = "IVF_FLAT"
    IVF_SQ8 = "IVF_SQ8"
    IVF_PQ = "IVF_PQ"
    HNSW = "HNSW"
    ANNOY = "ANNOY"
    RHNSW_FLAT = "RHNSW_FLAT"
    RHNSW_PQ = "RHNSW_PQ"
    RHNSW_SQ = "RHNSW_SQ"

    # Binary Vectors
    BIN_FLAT = "BIN_FLAT"
    BIN_IVF_FLAT = "BIN_IVF_FLAT"


@unique
class VectorSpace(str, Enum):
    # for floating point vectors:
    L2 = "L2"
    IP = "IP"
    # for binary vectors:
    JACCARD = "JACCARD"
    TANIMOTO = "TANIMOTO"
    HAMMING = "HAMMING"
    SUPERSTRUCTURE = "SUPERSTRUCTURE"
    SUBSTRUCTURE = "SUBSTRUCTURE"

@unique
class HashAlgo(str, Enum):
    md5 = "md5"
    sha1 = "sha1"
    sha256 = "sha256"
    sha512 = "sha512"
    sha3_224 = "sha3_224"
    sha3_256 = "sha3_256"
    sha3_384 = "sha3_384"
    sha3_512 = "sha3_512"
    # TODO: Require length
    # shake_128 = "shake_128 "
    # shake_256 = "shake_256"
    blake2b = "blake2b"
    blake2s = "blake2s"



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.hit_result import HitResult


T = TypeVar("T", bound="Hit")


@_attrs_define
class Hit:
    """
    Attributes:
        id (Union[int, str]):
        distance (float):
        result (HitResult):
        object_type (Union[Literal['Hit'], Unset]):  Default: 'Hit'.
    """

    id: Union[int, str]
    distance: float
    result: "HitResult"
    object_type: Union[Literal["Hit"], Unset] = "Hit"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id: Union[int, str]
        id = self.id

        distance = self.distance

        result = self.result.to_dict()

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "distance": distance,
                "result": result,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.hit_result import HitResult

        d = src_dict.copy()

        def _parse_id(data: object) -> Union[int, str]:
            return cast(Union[int, str], data)

        id = _parse_id(d.pop("id"))

        distance = d.pop("distance")

        result = HitResult.from_dict(d.pop("result"))

        object_type = d.pop("object_type", UNSET)

        hit = cls(
            id=id,
            distance=distance,
            result=result,
            object_type=object_type,
        )

        hit.additional_properties = d
        return hit

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" A client library (SDK) for accessing Lectric a Vector Database service """
from .client import AuthenticatedClient, Client
from .models import *
from .api import *
from .types import *
from .lectric_client import LectricClient
from .lectric_types import *
import os


with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "VERSION")) as version_fh:
  __version__ = version_fh.read().strip()

def version():
    return __version__



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.body_put_entry_by_file_exact_hash_collection_name_file_put import (
    BodyPutEntryByFileExactHashCollectionNameFilePut,
)
from ...models.http_validation_error import HTTPValidationError
from ...types import UNSET, Response, Unset


def _get_kwargs(
    collection_name: str,
    *,
    body: BodyPutEntryByFileExactHashCollectionNameFilePut,
    foreign_key: Union[None, Unset, int, str] = UNSET,
    metadata: Union[None, Unset, str] = UNSET,
    upsert: Union[None, Unset, bool] = False,
    timestamp: Union[None, Unset, str] = UNSET,
    store_raw_data: Union[None, Unset, bool] = True,
    ingestor: Union[None, Unset, str] = UNSET,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    params: Dict[str, Any] = {}

    json_foreign_key: Union[None, Unset, int, str]
    if isinstance(foreign_key, Unset):
        json_foreign_key = UNSET
    else:
        json_foreign_key = foreign_key
    params["foreign_key"] = json_foreign_key

    json_metadata: Union[None, Unset, str]
    if isinstance(metadata, Unset):
        json_metadata = UNSET
    else:
        json_metadata = metadata
    params["metadata"] = json_metadata

    json_upsert: Union[None, Unset, bool]
    if isinstance(upsert, Unset):
        json_upsert = UNSET
    else:
        json_upsert = upsert
    params["upsert"] = json_upsert

    json_timestamp: Union[None, Unset, str]
    if isinstance(timestamp, Unset):
        json_timestamp = UNSET
    else:
        json_timestamp = timestamp
    params["timestamp"] = json_timestamp

    json_store_raw_data: Union[None, Unset, bool]
    if isinstance(store_raw_data, Unset):
        json_store_raw_data = UNSET
    else:
        json_store_raw_data = store_raw_data
    params["store_raw_data"] = json_store_raw_data

    json_ingestor: Union[None, Unset, str]
    if isinstance(ingestor, Unset):
        json_ingestor = UNSET
    else:
        json_ingestor = ingestor
    params["ingestor"] = json_ingestor

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "put",
        "url": f"/exact/hash/{collection_name}/file",
        "params": params,
    }

    _body = body.to_multipart()

    _kwargs["files"] = _body

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, HTTPValidationError]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = response.json()
        return response_200
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = cast(Any, None)
        return response_404
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, HTTPValidationError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    collection_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyPutEntryByFileExactHashCollectionNameFilePut,
    foreign_key: Union[None, Unset, int, str] = UNSET,
    metadata: Union[None, Unset, str] = UNSET,
    upsert: Union[None, Unset, bool] = False,
    timestamp: Union[None, Unset, str] = UNSET,
    store_raw_data: Union[None, Unset, bool] = True,
    ingestor: Union[None, Unset, str] = UNSET,
) -> Response[Union[Any, HTTPValidationError]]:
    """Put Entry By File

    Args:
        collection_name (str):
        foreign_key (Union[None, Unset, int, str]):
        metadata (Union[None, Unset, str]):
        upsert (Union[None, Unset, bool]):  Default: False.
        timestamp (Union[None, Unset, str]):
        store_raw_data (Union[None, Unset, bool]):  Default: True.
        ingestor (Union[None, Unset, str]):
        body (BodyPutEntryByFileExactHashCollectionNameFilePut):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, HTTPValidationError]]
    """

    kwargs = _get_kwargs(
        collection_name=collection_name,
        body=body,
        foreign_key=foreign_key,
        metadata=metadata,
        upsert=upsert,
        timestamp=timestamp,
        store_raw_data=store_raw_data,
        ingestor=ingestor,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    collection_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyPutEntryByFileExactHashCollectionNameFilePut,
    foreign_key: Union[None, Unset, int, str] = UNSET,
    metadata: Union[None, Unset, str] = UNSET,
    upsert: Union[None, Unset, bool] = False,
    timestamp: Union[None, Unset, str] = UNSET,
    store_raw_data: Union[None, Unset, bool] = True,
    ingestor: Union[None, Unset, str] = UNSET,
) -> Optional[Union[Any, HTTPValidationError]]:
    """Put Entry By File

    Args:
        collection_name (str):
        foreign_key (Union[None, Unset, int, str]):
        metadata (Union[None, Unset, str]):
        upsert (Union[None, Unset, bool]):  Default: False.
        timestamp (Union[None, Unset, str]):
        store_raw_data (Union[None, Unset, bool]):  Default: True.
        ingestor (Union[None, Unset, str]):
        body (BodyPutEntryByFileExactHashCollectionNameFilePut):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, HTTPValidationError]
    """

    return sync_detailed(
        collection_name=collection_name,
        client=client,
        body=body,
        foreign_key=foreign_key,
        metadata=metadata,
        upsert=upsert,
        timestamp=timestamp,
        store_raw_data=store_raw_data,
        ingestor=ingestor,
    ).parsed


async def asyncio_detailed(
    collection_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyPutEntryByFileExactHashCollectionNameFilePut,
    foreign_key: Union[None, Unset, int, str] = UNSET,
    metadata: Union[None, Unset, str] = UNSET,
    upsert: Union[None, Unset, bool] = False,
    timestamp: Union[None, Unset, str] = UNSET,
    store_raw_data: Union[None, Unset, bool] = True,
    ingestor: Union[None, Unset, str] = UNSET,
) -> Response[Union[Any, HTTPValidationError]]:
    """Put Entry By File

    Args:
        collection_name (str):
        foreign_key (Union[None, Unset, int, str]):
        metadata (Union[None, Unset, str]):
        upsert (Union[None, Unset, bool]):  Default: False.
        timestamp (Union[None, Unset, str]):
        store_raw_data (Union[None, Unset, bool]):  Default: True.
        ingestor (Union[None, Unset, str]):
        body (BodyPutEntryByFileExactHashCollectionNameFilePut):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, HTTPValidationError]]
    """

    kwargs = _get_kwargs(
        collection_name=collection_name,
        body=body,
        foreign_key=foreign_key,
        metadata=metadata,
        upsert=upsert,
        timestamp=timestamp,
        store_raw_data=store_raw_data,
        ingestor=ingestor,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    collection_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyPutEntryByFileExactHashCollectionNameFilePut,
    foreign_key: Union[None, Unset, int, str] = UNSET,
    metadata: Union[None, Unset, str] = UNSET,
    upsert: Union[None, Unset, bool] = False,
    timestamp: Union[None, Unset, str] = UNSET,
    store_raw_data: Union[None, Unset, bool] = True,
    ingestor: Union[None, Unset, str] = UNSET,
) -> Optional[Union[Any, HTTPValidationError]]:
    """Put Entry By File

    Args:
        collection_name (str):
        foreign_key (Union[None, Unset, int, str]):
        metadata (Union[None, Unset, str]):
        upsert (Union[None, Unset, bool]):  Default: False.
        timestamp (Union[None, Unset, str]):
        store_raw_data (Union[None, Unset, bool]):  Default: True.
        ingestor (Union[None, Unset, str]):
        body (BodyPutEntryByFileExactHashCollectionNameFilePut):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, HTTPValidationError]
    """

    return (
        await asyncio_detailed(
            collection_name=collection_name,
            client=client,
            body=body,
            foreign_key=foreign_key,
            metadata=metadata,
            upsert=upsert,
            timestamp=timestamp,
            store_raw_data=store_raw_data,
            ingestor=ingestor,
        )
    ).parsed



from typing import Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="QueryMetaParams")


@_attrs_define
class QueryMetaParams:
    """
    Attributes:
        nprobe (int): Number of units to query. CPU: [1, nlist], GPU: [1, min(2048, nlist)
        object_type (Union[Literal['QueryMetaParams'], Unset]):  Default: 'QueryMetaParams'.
        ef (Union[None, Unset, int]): Search Scope. Range [top_k, 32768]
        search_k (Union[None, Unset, int]): The number of nodes to search. -1 means 5% of the whole data. Range {-1} U
            [top_k, n x n_trees]
    """

    nprobe: int
    object_type: Union[Literal["QueryMetaParams"], Unset] = "QueryMetaParams"
    ef: Union[None, Unset, int] = UNSET
    search_k: Union[None, Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        nprobe = self.nprobe

        object_type = self.object_type

        ef: Union[None, Unset, int]
        if isinstance(self.ef, Unset):
            ef = UNSET
        else:
            ef = self.ef

        search_k: Union[None, Unset, int]
        if isinstance(self.search_k, Unset):
            search_k = UNSET
        else:
            search_k = self.search_k

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "nprobe": nprobe,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if ef is not UNSET:
            field_dict["ef"] = ef
        if search_k is not UNSET:
            field_dict["search_k"] = search_k

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        nprobe = d.pop("nprobe")

        object_type = d.pop("object_type", UNSET)

        def _parse_ef(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        ef = _parse_ef(d.pop("ef", UNSET))

        def _parse_search_k(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        search_k = _parse_search_k(d.pop("search_k", UNSET))

        query_meta_params = cls(
            nprobe=nprobe,
            object_type=object_type,
            ef=ef,
            search_k=search_k,
        )

        query_meta_params.additional_properties = d
        return query_meta_params

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .body_hard_delete_entries_by_file_exact_entries_by_file_hard_delete_delete import (
    BodyHardDeleteEntriesByFileExactEntriesByFileHardDeleteDelete,
)
from .body_hard_delete_entries_exact_entries_hard_delete_delete import BodyHardDeleteEntriesExactEntriesHardDeleteDelete
from .body_ingest_w_file_ingest_file_post import BodyIngestWFileIngestFilePost
from .body_lookup_by_file_exact_lookup_file_collection_name_post import (
    BodyLookupByFileExactLookupFileCollectionNamePost,
)
from .body_put_entry_by_file_exact_hash_collection_name_file_put import BodyPutEntryByFileExactHashCollectionNameFilePut
from .body_soft_delete_entries_by_file_exact_entries_by_file_soft_delete_delete import (
    BodySoftDeleteEntriesByFileExactEntriesByFileSoftDeleteDelete,
)
from .body_soft_delete_entries_exact_entries_soft_delete_delete import BodySoftDeleteEntriesExactEntriesSoftDeleteDelete
from .collection import Collection
from .collection_in_spec import CollectionInSpec
from .collection_metadata import CollectionMetadata
from .collection_schema import CollectionSchema
from .exact_collection_entry import ExactCollectionEntry
from .exact_collection_entry_metadata import ExactCollectionEntryMetadata
from .field_schema import FieldSchema
from .hit import Hit
from .hit_result import HitResult
from .http_validation_error import HTTPValidationError
from .index import Index
from .index_in_spec import IndexInSpec
from .index_params import IndexParams
from .input_data import InputData
from .query_meta_params import QueryMetaParams
from .query_params import QueryParams
from .query_response import QueryResponse
from .query_spec import QuerySpec
from .schema_info_exact_schema_info_get_response_schema_info_exact_schema_info_get import (
    SchemaInfoExactSchemaInfoGetResponseSchemaInfoExactSchemaInfoGet,
)
from .select_entries_exact_select_post_response_200_item import SelectEntriesExactSelectPostResponse200Item
from .trash_can_entry import TrashCanEntry
from .validation_error import ValidationError
from .vector_query_spec import VectorQuerySpec

__all__ = (
    "BodyHardDeleteEntriesByFileExactEntriesByFileHardDeleteDelete",
    "BodyHardDeleteEntriesExactEntriesHardDeleteDelete",
    "BodyIngestWFileIngestFilePost",
    "BodyLookupByFileExactLookupFileCollectionNamePost",
    "BodyPutEntryByFileExactHashCollectionNameFilePut",
    "BodySoftDeleteEntriesByFileExactEntriesByFileSoftDeleteDelete",
    "BodySoftDeleteEntriesExactEntriesSoftDeleteDelete",
    "Collection",
    "CollectionInSpec",
    "CollectionMetadata",
    "CollectionSchema",
    "ExactCollectionEntry",
    "ExactCollectionEntryMetadata",
    "FieldSchema",
    "Hit",
    "HitResult",
    "HTTPValidationError",
    "Index",
    "IndexInSpec",
    "IndexParams",
    "InputData",
    "QueryMetaParams",
    "QueryParams",
    "QueryResponse",
    "QuerySpec",
    "SchemaInfoExactSchemaInfoGetResponseSchemaInfoExactSchemaInfoGet",
    "SelectEntriesExactSelectPostResponse200Item",
    "TrashCanEntry",
    "ValidationError",
    "VectorQuerySpec",
)



import pytest
import random
from typing import Dict, List
from .utils import (
    del_collection_if_exists,
    gen_random_fp_collection_spec,
    gen_random_name,
    create_collection, gen_random_fp_vectors)
import lectric as lc

DIM = 8
NUM_ELEMS = 20

METRIC_TYPE = lc.VectorSpace.L2

@pytest.fixture(scope="module")
def collection(client: lc.LectricClient) -> lc.Collection:
    coll: lc.Collection = create_collection(client, gen_random_fp_collection_spec(8, random_field=True))

    # Ingest some data
    input_data = lc.InputData(
        collection_name=coll.name,
        data=[gen_random_fp_vectors(DIM, NUM_ELEMS),
                [float(random.randrange(-20, -10)) for _ in range(NUM_ELEMS)]
            ]
    )

    client.ingest(input_data)
    client.create_index(lc.IndexInSpec(
        coll.name, "vecs", index=lc.Index(
            lc.IndexType.IVF_FLAT, metric_type=METRIC_TYPE, params=lc.IndexParams()
        )
    ))

    yield coll

    del_collection_if_exists(coll.name, client)

def test_query_vectors(client: lc.LectricClient, collection: lc.Collection):
    K = 10
    spec: lc.VectorQuerySpec = lc.VectorQuerySpec(
        data=gen_random_fp_vectors(DIM, 2),
        collection_name=collection.name,
        search_field="vecs",
        search_params=lc.QueryParams(metric_type=METRIC_TYPE,
                                params=lc.QueryMetaParams(
                                nprobe=5
                            )),
        output_fields=["id", "random"], # NOTE: You cannot get vector field back as output_field
        limit=K
        )

    qres: lc.QueryResponse = client.query(spec)
    assert len(qres.hits) == 2

    for hit in qres.hits:
        assert len(hit) == K

    # Assert our output fields are there
    assert "id" in qres.hits[0][0].result
    assert "random" in qres.hits[0][0].result

def test_query_fields(client: lc.LectricClient, collection: lc.Collection):
    spec: lc.QuerySpec = lc.QuerySpec(
        expr="random > -14",
        collection_name=collection.name,
        output_fields=["random", "vecs"]
    )

    qres = client.query(spec)

    assert "id" in qres[0] # ID will always be in the response
    assert "vecs" in qres[0]
    assert "random" in qres[0]
    assert len(qres[0]["vecs"]) == DIM


def test_varchar_field(client: lc.LectricClient):
    coll_dim = 8

    fields: List[lc.FieldSchema] = [
        lc.FieldSchema(name="id", dtype=lc.DataType.INT64, is_primary=True,
            auto_id=True, description="Primary key"),
        lc.FieldSchema(name="vecs", dtype=lc.DataType.FLOAT_VECTOR,
                        dim=coll_dim, description="floating point embeddings"),
        lc.FieldSchema(name="foreign_id", dtype=lc.DataType.VARCHAR, max_length=10,
                       description="foreign ID key", is_primary=False),
    ]

    schema: lc.CollectionSchema = lc.CollectionSchema(fields=fields, description="Test collection schema")

    collection: lc.Collection = None
    try:
        collection = client.create_collection(lc.CollectionInSpec(name=gen_random_name(10), coll_schema=schema))

        # Ingest
        vecs: List[List[float]] = gen_random_fp_vectors(coll_dim, 3)
        fids: List[str] = [gen_random_name(6) for i in range(len(vecs))]

        data = [vecs, fids]
        client.ingest(lc.InputData(collection_name=collection.name, data=data))

        # FIXME: Add index
        client.create_index(lc.IndexInSpec(
            collection.name, "vecs", index=lc.Index(
                lc.IndexType.IVF_FLAT, metric_type=METRIC_TYPE, params=lc.IndexParams()
            )
        ))

        # Query for string prop back (emulate foreign key lookup)
        for fid in fids:
            qres: List[Dict] = client.query(lc.QuerySpec(collection_name=collection.name,
                    expr=f"foreign_id == \"{fid}\"", output_fields=["foreign_id", "vecs"]))

            assert qres[0]["foreign_id"] == fid

    finally:
        if collection:
            del_collection_if_exists(collection_name=collection.name, client=client)



import datetime
from typing import Any, Dict, List, Literal, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="TrashCanEntry")


@_attrs_define
class TrashCanEntry:
    """
    Attributes:
        collection (str):
        data_collection_primary_key (str):
        soft_deletion_timestamp (datetime.datetime):
        soft_deleter (str):
        link (str):
        object_type (Union[Literal['TrashCanEntry'], Unset]):  Default: 'TrashCanEntry'.
    """

    collection: str
    data_collection_primary_key: str
    soft_deletion_timestamp: datetime.datetime
    soft_deleter: str
    link: str
    object_type: Union[Literal["TrashCanEntry"], Unset] = "TrashCanEntry"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        collection = self.collection

        data_collection_primary_key = self.data_collection_primary_key

        soft_deletion_timestamp = self.soft_deletion_timestamp.isoformat()

        soft_deleter = self.soft_deleter

        link = self.link

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "collection": collection,
                "data_collection_primary_key": data_collection_primary_key,
                "soft_deletion_timestamp": soft_deletion_timestamp,
                "soft_deleter": soft_deleter,
                "link": link,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        collection = d.pop("collection")

        data_collection_primary_key = d.pop("data_collection_primary_key")

        soft_deletion_timestamp = isoparse(d.pop("soft_deletion_timestamp"))

        soft_deleter = d.pop("soft_deleter")

        link = d.pop("link")

        object_type = d.pop("object_type", UNSET)

        trash_can_entry = cls(
            collection=collection,
            data_collection_primary_key=data_collection_primary_key,
            soft_deletion_timestamp=soft_deletion_timestamp,
            soft_deleter=soft_deleter,
            link=link,
            object_type=object_type,
        )

        trash_can_entry.additional_properties = d
        return trash_can_entry

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
import tempfile
import requests
import numpy as np
from typing import BinaryIO, List, Any, Union
import io, os
from .models.collection_schema import CollectionSchema
from .models.field_schema import FieldSchema
from .lectric_types import DataType


def create_random_vectors(nvects: int, dim: int) -> List[List[float]]:
    return [np.random.rand(dim).tolist() for _ in range(nvects)]

def check_ok(resp: requests.Response):
    if resp.status_code != HTTPStatus.OK:
        raise RuntimeError(resp.json())

def is_file_like(obj: Any) -> bool:
    """Does the object appear to be a file?

    Args:
        obj (Any): The object in question

    Returns:
        [bool]: True if is file like else False
    """
    return isinstance(obj, io.TextIOBase) or \
                isinstance(obj, io.BufferedIOBase) or \
                isinstance(obj, io.RawIOBase) or \
                isinstance(obj, io.IOBase)


def to_tmpfile(_input: bytes) -> BinaryIO:
    """Write a sequence of bytes a temporary file

    Args:
        _input (bytes): The data to be written to temp file

    Returns:
        [BinaryIO]: The temporary file that is closed when garbage collectede
    """
    fp = tempfile.TemporaryFile()
    fp.write(_input)
    fp.seek(0)
    return fp


def get_tmp_file(file: Union[str, bytes, BinaryIO]) -> BinaryIO:
    if is_file_like(file):
        tmp_file = file
    elif isinstance(file, str):
        if not os.path.exists(file):
            raise RuntimeError("`file` arg as str must be the path to file")
        tmp_file = open(file, "rb")

    elif isinstance(file, bytes): # Assume encoded JPG
        tmp_file = to_tmpfile(file)
    else:
        raise RuntimeError("Unable to process `file` argument, unsupported input type")

    return tmp_file

def pformat_field_schema(fschema: FieldSchema) -> str:
    desc = "" if not fschema.description else f"Description: {fschema.description}"
    dim = "" if not fschema.dim else f"Dim: {fschema.dim}"
    max_len = "" if not fschema.max_length else f"Max length: {fschema.max_length}"
    s =  f"""
            FieldSchema:
                Name: {fschema.name}
                {desc}
                Type: {DataType(fschema.dtype)}
                Primary key: {fschema.is_primary}
                Auto ID: {bool(fschema.auto_id)}
                {dim}
                {max_len}
        """.rstrip()

    return s

def pformat_collection_schema(cschema: CollectionSchema) -> str:
    schema_str: str = "\nCollectionSchema:"
    schema_str += "" if not cschema.description else f"\n\tDescription: {cschema.description}"
    for field in cschema.fields:
        field_str = pformat_field_schema(field)
        schema_str += field_str
    return schema_str



import datetime
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.exact_collection_entry import ExactCollectionEntry
from ...models.http_validation_error import HTTPValidationError
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    collection_name: str,
    ingestor: Union[Unset, str] = UNSET,
    start_date: Union[None, Unset, datetime.date] = UNSET,
    end_date: Union[None, Unset, datetime.date] = UNSET,
    ingest_source: Union[None, Unset, str] = UNSET,
    traits: Union[None, Unset, str] = UNSET,
    order_by: Union[None, Unset, str] = UNSET,
    desc: Union[Unset, bool] = True,
    limit: Union[Unset, int] = 1000,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    params["collection_name"] = collection_name

    params["ingestor"] = ingestor

    json_start_date: Union[None, Unset, str]
    if isinstance(start_date, Unset):
        json_start_date = UNSET
    elif isinstance(start_date, datetime.date):
        json_start_date = start_date.isoformat()
    else:
        json_start_date = start_date
    params["start_date"] = json_start_date

    json_end_date: Union[None, Unset, str]
    if isinstance(end_date, Unset):
        json_end_date = UNSET
    elif isinstance(end_date, datetime.date):
        json_end_date = end_date.isoformat()
    else:
        json_end_date = end_date
    params["end_date"] = json_end_date

    json_ingest_source: Union[None, Unset, str]
    if isinstance(ingest_source, Unset):
        json_ingest_source = UNSET
    else:
        json_ingest_source = ingest_source
    params["ingest_source"] = json_ingest_source

    json_traits: Union[None, Unset, str]
    if isinstance(traits, Unset):
        json_traits = UNSET
    else:
        json_traits = traits
    params["traits"] = json_traits

    json_order_by: Union[None, Unset, str]
    if isinstance(order_by, Unset):
        json_order_by = UNSET
    else:
        json_order_by = order_by
    params["order_by"] = json_order_by

    params["desc"] = desc

    params["limit"] = limit

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "get",
        "url": "/exact/collection_entries",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = ExactCollectionEntry.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = cast(Any, None)
        return response_404
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    collection_name: str,
    ingestor: Union[Unset, str] = UNSET,
    start_date: Union[None, Unset, datetime.date] = UNSET,
    end_date: Union[None, Unset, datetime.date] = UNSET,
    ingest_source: Union[None, Unset, str] = UNSET,
    traits: Union[None, Unset, str] = UNSET,
    order_by: Union[None, Unset, str] = UNSET,
    desc: Union[Unset, bool] = True,
    limit: Union[Unset, int] = 1000,
) -> Response[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    """Get Exact Collection Entries

     Get entries from the collection

    Args:
        collection_name (str):
        ingestor (Union[Unset, str]):
        start_date (Union[None, Unset, datetime.date]):
        end_date (Union[None, Unset, datetime.date]):
        ingest_source (Union[None, Unset, str]):
        traits (Union[None, Unset, str]):
        order_by (Union[None, Unset, str]):
        desc (Union[Unset, bool]):  Default: True.
        limit (Union[Unset, int]):  Default: 1000.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, HTTPValidationError, List['ExactCollectionEntry']]]
    """

    kwargs = _get_kwargs(
        collection_name=collection_name,
        ingestor=ingestor,
        start_date=start_date,
        end_date=end_date,
        ingest_source=ingest_source,
        traits=traits,
        order_by=order_by,
        desc=desc,
        limit=limit,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
    collection_name: str,
    ingestor: Union[Unset, str] = UNSET,
    start_date: Union[None, Unset, datetime.date] = UNSET,
    end_date: Union[None, Unset, datetime.date] = UNSET,
    ingest_source: Union[None, Unset, str] = UNSET,
    traits: Union[None, Unset, str] = UNSET,
    order_by: Union[None, Unset, str] = UNSET,
    desc: Union[Unset, bool] = True,
    limit: Union[Unset, int] = 1000,
) -> Optional[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    """Get Exact Collection Entries

     Get entries from the collection

    Args:
        collection_name (str):
        ingestor (Union[Unset, str]):
        start_date (Union[None, Unset, datetime.date]):
        end_date (Union[None, Unset, datetime.date]):
        ingest_source (Union[None, Unset, str]):
        traits (Union[None, Unset, str]):
        order_by (Union[None, Unset, str]):
        desc (Union[Unset, bool]):  Default: True.
        limit (Union[Unset, int]):  Default: 1000.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, HTTPValidationError, List['ExactCollectionEntry']]
    """

    return sync_detailed(
        client=client,
        collection_name=collection_name,
        ingestor=ingestor,
        start_date=start_date,
        end_date=end_date,
        ingest_source=ingest_source,
        traits=traits,
        order_by=order_by,
        desc=desc,
        limit=limit,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    collection_name: str,
    ingestor: Union[Unset, str] = UNSET,
    start_date: Union[None, Unset, datetime.date] = UNSET,
    end_date: Union[None, Unset, datetime.date] = UNSET,
    ingest_source: Union[None, Unset, str] = UNSET,
    traits: Union[None, Unset, str] = UNSET,
    order_by: Union[None, Unset, str] = UNSET,
    desc: Union[Unset, bool] = True,
    limit: Union[Unset, int] = 1000,
) -> Response[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    """Get Exact Collection Entries

     Get entries from the collection

    Args:
        collection_name (str):
        ingestor (Union[Unset, str]):
        start_date (Union[None, Unset, datetime.date]):
        end_date (Union[None, Unset, datetime.date]):
        ingest_source (Union[None, Unset, str]):
        traits (Union[None, Unset, str]):
        order_by (Union[None, Unset, str]):
        desc (Union[Unset, bool]):  Default: True.
        limit (Union[Unset, int]):  Default: 1000.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, HTTPValidationError, List['ExactCollectionEntry']]]
    """

    kwargs = _get_kwargs(
        collection_name=collection_name,
        ingestor=ingestor,
        start_date=start_date,
        end_date=end_date,
        ingest_source=ingest_source,
        traits=traits,
        order_by=order_by,
        desc=desc,
        limit=limit,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    collection_name: str,
    ingestor: Union[Unset, str] = UNSET,
    start_date: Union[None, Unset, datetime.date] = UNSET,
    end_date: Union[None, Unset, datetime.date] = UNSET,
    ingest_source: Union[None, Unset, str] = UNSET,
    traits: Union[None, Unset, str] = UNSET,
    order_by: Union[None, Unset, str] = UNSET,
    desc: Union[Unset, bool] = True,
    limit: Union[Unset, int] = 1000,
) -> Optional[Union[Any, HTTPValidationError, List["ExactCollectionEntry"]]]:
    """Get Exact Collection Entries

     Get entries from the collection

    Args:
        collection_name (str):
        ingestor (Union[Unset, str]):
        start_date (Union[None, Unset, datetime.date]):
        end_date (Union[None, Unset, datetime.date]):
        ingest_source (Union[None, Unset, str]):
        traits (Union[None, Unset, str]):
        order_by (Union[None, Unset, str]):
        desc (Union[Unset, bool]):  Default: True.
        limit (Union[Unset, int]):  Default: 1000.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, HTTPValidationError, List['ExactCollectionEntry']]
    """

    return (
        await asyncio_detailed(
            client=client,
            collection_name=collection_name,
            ingestor=ingestor,
            start_date=start_date,
            end_date=end_date,
            ingest_source=ingest_source,
            traits=traits,
            order_by=order_by,
            desc=desc,
            limit=limit,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.collection_schema import CollectionSchema


T = TypeVar("T", bound="CollectionInSpec")


@_attrs_define
class CollectionInSpec:
    """
    Attributes:
        name (str):
        coll_schema (CollectionSchema):
        object_type (Union[Literal['CollectionInSpec'], Unset]):  Default: 'CollectionInSpec'.
        consistency_level (Union[Unset, str]):  Default: 'Session'.
        approx (Union[None, Unset, bool]):  Default: True.
    """

    name: str
    coll_schema: "CollectionSchema"
    object_type: Union[Literal["CollectionInSpec"], Unset] = "CollectionInSpec"
    consistency_level: Union[Unset, str] = "Session"
    approx: Union[None, Unset, bool] = True
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        coll_schema = self.coll_schema.to_dict()

        object_type = self.object_type

        consistency_level = self.consistency_level

        approx: Union[None, Unset, bool]
        if isinstance(self.approx, Unset):
            approx = UNSET
        else:
            approx = self.approx

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "coll_schema": coll_schema,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if consistency_level is not UNSET:
            field_dict["consistency_level"] = consistency_level
        if approx is not UNSET:
            field_dict["approx"] = approx

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.collection_schema import CollectionSchema

        d = src_dict.copy()
        name = d.pop("name")

        coll_schema = CollectionSchema.from_dict(d.pop("coll_schema"))

        object_type = d.pop("object_type", UNSET)

        consistency_level = d.pop("consistency_level", UNSET)

        def _parse_approx(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        approx = _parse_approx(d.pop("approx", UNSET))

        collection_in_spec = cls(
            name=name,
            coll_schema=coll_schema,
            object_type=object_type,
            consistency_level=consistency_level,
            approx=approx,
        )

        collection_in_spec.additional_properties = d
        return collection_in_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import setuptools
from setuptools import setup


with open("lectric/VERSION") as version_fh:
  version = version_fh.read().strip()

with open("README.md") as readme_fh:
  long_description = readme_fh.read()


setup(
    name='lectric-sdk',
    packages=setuptools.find_packages(exclude=["tests"]),
    include_package_data=True,
    version=version,
    description='Lectric client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Disa Mhembere',
    author_email='disamhembere@microsoft.com',
    keywords=['lectric', 'vdb', 'vector database'],
    classifiers=["Programming Language :: Python :: 3",
                 "Operating System :: OS Independent"],
    install_requires=[
        "setuptools>=41.0.1",
        "httpx>=0.23.0",
        "attrs>=21.4.0",
        "wheel",
        "requests",
        "numpy",
        "python-dateutil"
    ],
    python_requires=">=3.7,<3.10",
    zip_safe=False
)



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.query_meta_params import QueryMetaParams


T = TypeVar("T", bound="QueryParams")


@_attrs_define
class QueryParams:
    """
    Attributes:
        metric_type (str): A string corresponding to the Enum: VectorSpace
        params (QueryMetaParams):
        object_type (Union[Literal['QueryParams'], Unset]):  Default: 'QueryParams'.
    """

    metric_type: str
    params: "QueryMetaParams"
    object_type: Union[Literal["QueryParams"], Unset] = "QueryParams"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        metric_type = self.metric_type

        params = self.params.to_dict()

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "metric_type": metric_type,
                "params": params,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.query_meta_params import QueryMetaParams

        d = src_dict.copy()
        metric_type = d.pop("metric_type")

        params = QueryMetaParams.from_dict(d.pop("params"))

        object_type = d.pop("object_type", UNSET)

        query_params = cls(
            metric_type=metric_type,
            params=params,
            object_type=object_type,
        )

        query_params.additional_properties = d
        return query_params

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.query_params import QueryParams


T = TypeVar("T", bound="VectorQuerySpec")


@_attrs_define
class VectorQuerySpec:
    """
    Attributes:
        data (List[List[Any]]):
        collection_name (str):
        search_field (str):
        search_params (QueryParams):
        output_fields (List[str]):
        limit (int):
        object_type (Union[Literal['VectorQuerySpec'], Unset]):  Default: 'VectorQuerySpec'.
        expr (Union[None, Unset, str]):
    """

    data: List[List[Any]]
    collection_name: str
    search_field: str
    search_params: "QueryParams"
    output_fields: List[str]
    limit: int
    object_type: Union[Literal["VectorQuerySpec"], Unset] = "VectorQuerySpec"
    expr: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        data = []
        for data_item_data in self.data:
            data_item = data_item_data

            data.append(data_item)

        collection_name = self.collection_name

        search_field = self.search_field

        search_params = self.search_params.to_dict()

        output_fields = self.output_fields

        limit = self.limit

        object_type = self.object_type

        expr: Union[None, Unset, str]
        if isinstance(self.expr, Unset):
            expr = UNSET
        else:
            expr = self.expr

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "data": data,
                "collection_name": collection_name,
                "search_field": search_field,
                "search_params": search_params,
                "output_fields": output_fields,
                "limit": limit,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if expr is not UNSET:
            field_dict["expr"] = expr

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.query_params import QueryParams

        d = src_dict.copy()
        data = []
        _data = d.pop("data")
        for data_item_data in _data:
            data_item = cast(List[Any], data_item_data)

            data.append(data_item)

        collection_name = d.pop("collection_name")

        search_field = d.pop("search_field")

        search_params = QueryParams.from_dict(d.pop("search_params"))

        output_fields = cast(List[str], d.pop("output_fields"))

        limit = d.pop("limit")

        object_type = d.pop("object_type", UNSET)

        def _parse_expr(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        expr = _parse_expr(d.pop("expr", UNSET))

        vector_query_spec = cls(
            data=data,
            collection_name=collection_name,
            search_field=search_field,
            search_params=search_params,
            output_fields=output_fields,
            limit=limit,
            object_type=object_type,
            expr=expr,
        )

        vector_query_spec.additional_properties = d
        return vector_query_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="FieldSchema")


@_attrs_define
class FieldSchema:
    """
    Attributes:
        name (str):
        dtype (int):
        object_type (Union[Literal['FieldSchema'], Unset]):  Default: 'FieldSchema'.
        is_primary (Union[None, Unset, bool]):  Default: False.
        auto_id (Union[None, Unset, bool]):  Default: False.
        description (Union[None, Unset, str]):  Default: ''.
        dim (Union[None, Unset, int]):
        max_length (Union[None, Unset, int]):
    """

    name: str
    dtype: int
    object_type: Union[Literal["FieldSchema"], Unset] = "FieldSchema"
    is_primary: Union[None, Unset, bool] = False
    auto_id: Union[None, Unset, bool] = False
    description: Union[None, Unset, str] = ""
    dim: Union[None, Unset, int] = UNSET
    max_length: Union[None, Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        dtype = self.dtype

        object_type = self.object_type

        is_primary: Union[None, Unset, bool]
        if isinstance(self.is_primary, Unset):
            is_primary = UNSET
        else:
            is_primary = self.is_primary

        auto_id: Union[None, Unset, bool]
        if isinstance(self.auto_id, Unset):
            auto_id = UNSET
        else:
            auto_id = self.auto_id

        description: Union[None, Unset, str]
        if isinstance(self.description, Unset):
            description = UNSET
        else:
            description = self.description

        dim: Union[None, Unset, int]
        if isinstance(self.dim, Unset):
            dim = UNSET
        else:
            dim = self.dim

        max_length: Union[None, Unset, int]
        if isinstance(self.max_length, Unset):
            max_length = UNSET
        else:
            max_length = self.max_length

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "dtype": dtype,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if is_primary is not UNSET:
            field_dict["is_primary"] = is_primary
        if auto_id is not UNSET:
            field_dict["auto_id"] = auto_id
        if description is not UNSET:
            field_dict["description"] = description
        if dim is not UNSET:
            field_dict["dim"] = dim
        if max_length is not UNSET:
            field_dict["max_length"] = max_length

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        dtype = d.pop("dtype")

        object_type = d.pop("object_type", UNSET)

        def _parse_is_primary(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        is_primary = _parse_is_primary(d.pop("is_primary", UNSET))

        def _parse_auto_id(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        auto_id = _parse_auto_id(d.pop("auto_id", UNSET))

        def _parse_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        description = _parse_description(d.pop("description", UNSET))

        def _parse_dim(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        dim = _parse_dim(d.pop("dim", UNSET))

        def _parse_max_length(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        max_length = _parse_max_length(d.pop("max_length", UNSET))

        field_schema = cls(
            name=name,
            dtype=dtype,
            object_type=object_type,
            is_primary=is_primary,
            auto_id=auto_id,
            description=description,
            dim=dim,
            max_length=max_length,
        )

        field_schema.additional_properties = d
        return field_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="QuerySpec")


@_attrs_define
class QuerySpec:
    """
    Attributes:
        expr (str):
        collection_name (str):
        object_type (Union[Literal['QuerySpec'], Unset]):  Default: 'QuerySpec'.
        output_fields (Union[List[str], None, Unset]):
    """

    expr: str
    collection_name: str
    object_type: Union[Literal["QuerySpec"], Unset] = "QuerySpec"
    output_fields: Union[List[str], None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        expr = self.expr

        collection_name = self.collection_name

        object_type = self.object_type

        output_fields: Union[List[str], None, Unset]
        if isinstance(self.output_fields, Unset):
            output_fields = UNSET
        elif isinstance(self.output_fields, list):
            output_fields = self.output_fields

        else:
            output_fields = self.output_fields

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "expr": expr,
                "collection_name": collection_name,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type
        if output_fields is not UNSET:
            field_dict["output_fields"] = output_fields

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        expr = d.pop("expr")

        collection_name = d.pop("collection_name")

        object_type = d.pop("object_type", UNSET)

        def _parse_output_fields(data: object) -> Union[List[str], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                output_fields_type_0 = cast(List[str], data)

                return output_fields_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None, Unset], data)

        output_fields = _parse_output_fields(d.pop("output_fields", UNSET))

        query_spec = cls(
            expr=expr,
            collection_name=collection_name,
            object_type=object_type,
            output_fields=output_fields,
        )

        query_spec.additional_properties = d
        return query_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.hit import Hit


T = TypeVar("T", bound="QueryResponse")


@_attrs_define
class QueryResponse:
    """
    Attributes:
        collection_name (str):
        hits (List[List['Hit']]):
        query_latency (float):
        total_latency (float):
        object_type (Union[Literal['QueryResponse'], Unset]):  Default: 'QueryResponse'.
    """

    collection_name: str
    hits: List[List["Hit"]]
    query_latency: float
    total_latency: float
    object_type: Union[Literal["QueryResponse"], Unset] = "QueryResponse"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        collection_name = self.collection_name

        hits = []
        for hits_item_data in self.hits:
            hits_item = []
            for hits_item_item_data in hits_item_data:
                hits_item_item = hits_item_item_data.to_dict()
                hits_item.append(hits_item_item)

            hits.append(hits_item)

        query_latency = self.query_latency

        total_latency = self.total_latency

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "collection_name": collection_name,
                "hits": hits,
                "query_latency": query_latency,
                "total_latency": total_latency,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.hit import Hit

        d = src_dict.copy()
        collection_name = d.pop("collection_name")

        hits = []
        _hits = d.pop("hits")
        for hits_item_data in _hits:
            hits_item = []
            _hits_item = hits_item_data
            for hits_item_item_data in _hits_item:
                hits_item_item = Hit.from_dict(hits_item_item_data)

                hits_item.append(hits_item_item)

            hits.append(hits_item)

        query_latency = d.pop("query_latency")

        total_latency = d.pop("total_latency")

        object_type = d.pop("object_type", UNSET)

        query_response = cls(
            collection_name=collection_name,
            hits=hits,
            query_latency=query_latency,
            total_latency=total_latency,
            object_type=object_type,
        )

        query_response.additional_properties = d
        return query_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.exact_collection_entry_metadata import ExactCollectionEntryMetadata


T = TypeVar("T", bound="ExactCollectionEntry")


@_attrs_define
class ExactCollectionEntry:
    """
    Attributes:
        id (str):
        link (str):
        timestamp (datetime.datetime):
        is_file_upload (bool):
        metadata (ExactCollectionEntryMetadata):
        is_marked_for_deletion (bool):
        ingestor (str):
        fk (str):
        object_type (Union[Literal['ExactCollectionEntry'], Unset]):  Default: 'ExactCollectionEntry'.
    """

    id: str
    link: str
    timestamp: datetime.datetime
    is_file_upload: bool
    metadata: "ExactCollectionEntryMetadata"
    is_marked_for_deletion: bool
    ingestor: str
    fk: str
    object_type: Union[Literal["ExactCollectionEntry"], Unset] = "ExactCollectionEntry"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        link = self.link

        timestamp = self.timestamp.isoformat()

        is_file_upload = self.is_file_upload

        metadata = self.metadata.to_dict()

        is_marked_for_deletion = self.is_marked_for_deletion

        ingestor = self.ingestor

        fk = self.fk

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "link": link,
                "timestamp": timestamp,
                "is_file_upload": is_file_upload,
                "metadata": metadata,
                "is_marked_for_deletion": is_marked_for_deletion,
                "ingestor": ingestor,
                "fk": fk,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.exact_collection_entry_metadata import ExactCollectionEntryMetadata

        d = src_dict.copy()
        id = d.pop("id")

        link = d.pop("link")

        timestamp = isoparse(d.pop("timestamp"))

        is_file_upload = d.pop("is_file_upload")

        metadata = ExactCollectionEntryMetadata.from_dict(d.pop("metadata"))

        is_marked_for_deletion = d.pop("is_marked_for_deletion")

        ingestor = d.pop("ingestor")

        fk = d.pop("fk")

        object_type = d.pop("object_type", UNSET)

        exact_collection_entry = cls(
            id=id,
            link=link,
            timestamp=timestamp,
            is_file_upload=is_file_upload,
            metadata=metadata,
            is_marked_for_deletion=is_marked_for_deletion,
            ingestor=ingestor,
            fk=fk,
            object_type=object_type,
        )

        exact_collection_entry.additional_properties = d
        return exact_collection_entry

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Literal, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CollectionMetadata")


@_attrs_define
class CollectionMetadata:
    """
    Attributes:
        name (str):
        description (str):
        hash_algo (str):
        etag (str):
        is_marked_for_deletion (bool):
        soft_deleter (str):
        object_type (Union[Literal['CollectionMetadata'], Unset]):  Default: 'CollectionMetadata'.
    """

    name: str
    description: str
    hash_algo: str
    etag: str
    is_marked_for_deletion: bool
    soft_deleter: str
    object_type: Union[Literal["CollectionMetadata"], Unset] = "CollectionMetadata"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        description = self.description

        hash_algo = self.hash_algo

        etag = self.etag

        is_marked_for_deletion = self.is_marked_for_deletion

        soft_deleter = self.soft_deleter

        object_type = self.object_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "description": description,
                "hash_algo": hash_algo,
                "etag": etag,
                "is_marked_for_deletion": is_marked_for_deletion,
                "soft_deleter": soft_deleter,
            }
        )
        if object_type is not UNSET:
            field_dict["object_type"] = object_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        description = d.pop("description")

        hash_algo = d.pop("hash_algo")

        etag = d.pop("etag")

        is_marked_for_deletion = d.pop("is_marked_for_deletion")

        soft_deleter = d.pop("soft_deleter")

        object_type = d.pop("object_type", UNSET)

        collection_metadata = cls(
            name=name,
            description=description,
            hash_algo=hash_algo,
            etag=etag,
            is_marked_for_deletion=is_marked_for_deletion,
            soft_deleter=soft_deleter,
            object_type=object_type,
        )

        collection_metadata.additional_properties = d
        return collection_metadata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
